#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include "transform.h"
#include "primitives.h"

#define sqr(x) ((x)*(x))
#define EPS 0.0000001
#define INFTY 10e9

char infilename[1000];
int VERBOSE=0,DEBUG=0;
int XSIZE=640,YSIZE=480;
int HLBKA_VNORENIA=4;
int TWOSIDED_LIGHTING=0;
FILE *f;

tCamera Kamera;
tSphere SphereList[100]; int SphereCount=0;
tTriangle TriangleList[10000]; int TriangleCount=0;
tVector AmbientLight;
tPointLight PointLights[100]; int PointLightCount=0;

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   POMOCNE HALUZE                                          //
///////////////////////////////////////////////////////////////////////////////////////////////

// ak sme verbose, vypise message 
void message(const char *format, ...){ va_list ap; va_start(ap,format); if (VERBOSE) vprintf(format,ap); va_end(ap); }
void dbgmessage(const char *format, ...){ va_list ap; va_start(ap,format); if (DEBUG) vprintf(format,ap); va_end(ap); }

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   NACITANIE                                               //
///////////////////////////////////////////////////////////////////////////////////////////////

void LoadCamera(void){
  char row[1000],*rest;
  int i,rl;
  tVector smer,perp;
  double pom;
 
  Kamera.location[0]=Kamera.location[1]=Kamera.location[2]=0; Kamera.location[3]=1;
  Kamera.direction[2]=1; Kamera.direction[0]=Kamera.direction[1]=Kamera.direction[3]=0;
  Kamera.right[0]=0.3*1.33333333; Kamera.right[1]=Kamera.right[2]=Kamera.right[3]=0;
  Kamera.up[1]=0.3*1; Kamera.up[0]=Kamera.up[2]=Kamera.up[3]=0;
  
  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: camera: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      if (!strcmp(row,"location")) {
        sscanf(rest,"%lf , %lf , %lf",&Kamera.location[0],&Kamera.location[1],&Kamera.location[2]);
        dbgmessage("Load: camera: location %.2f %.2f %.2f\n",Kamera.location[0],Kamera.location[1],Kamera.location[2]);
        break;
      }
      if (!strcmp(row,"look_at")) {
        sscanf(rest,"%lf , %lf , %lf",&smer[0],&smer[1],&smer[2]); smer[3]=0;
        for (i=0;i<3;i++) { smer[i]-=Kamera.location[i]; Kamera.direction[i]=perp[i]=smer[i]; }
        perp[1]=0;
        if (Length(perp)<EPS){
          // pozerame hore alebo dole
        } else {
          pom=perp[0]; perp[0]=perp[2]; perp[2]=-pom; // note the minus sign!
          CrossProduct(smer,perp,&Kamera.up); Normalize(&Kamera.up);
          CrossProduct(Kamera.up,Kamera.direction,&Kamera.right);
          for (i=0;i<3;i++) { 
            Kamera.up[i]*=0.3*Length(Kamera.direction);
            Kamera.right[i]*=0.4*Length(Kamera.direction);
          }
        }
        dbgmessage("Load: camera: direction %.2f %.2f %.2f\n",Kamera.direction[0],Kamera.direction[1],Kamera.direction[2]);
        dbgmessage("Load: camera: up %.2f %.2f %.2f\n",Kamera.up[0],Kamera.up[1],Kamera.up[2]);
        dbgmessage("Load: camera: right %.2f %.2f %.2f\n",Kamera.right[0],Kamera.right[1],Kamera.right[2]);
        break;
      }
      if (!strcmp(row,"up")) {
        sscanf(rest,"%lf , %lf , %lf",&Kamera.up[0],&Kamera.up[1],&Kamera.up[2]);
          CrossProduct(Kamera.up,Kamera.direction,&Kamera.right);
          Normalize(&Kamera.right); 
          for (i=0;i<3;i++) Kamera.right[i]*=4/3.0*Length(Kamera.up);
        break;
      }
      dbgmessage("Load: camera: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
}

void LoadSphere(void){
  char row[1000],*rest;
  int i,rl;
 
  // NASTAVIME DEFAULT HODNOTY FARBY A MATERIALU
  SphereList[SphereCount].ambientIntensity=0.15;
  SphereList[SphereCount].diffuseColor[0]=SphereList[SphereCount].diffuseColor[1]
    =SphereList[SphereCount].diffuseColor[2]=1.0;
  SphereList[SphereCount].specularColor[0]=SphereList[SphereCount].specularColor[1]
    =SphereList[SphereCount].specularColor[2]=0.0;
  SphereList[SphereCount].emissiveColor[0]=SphereList[SphereCount].emissiveColor[1]
    =SphereList[SphereCount].emissiveColor[2]=0.0;
  
  SphereList[SphereCount].reflection[AMBIENT]=0.15; 
  SphereList[SphereCount].reflection[DIFFUSE]=0.8; 
  SphereList[SphereCount].reflection[SPECULAR]=0.0; 
  SphereList[SphereCount].transparency=0.0; 
  SphereList[SphereCount].refractionindex=1.0; 
  SphereList[SphereCount].reflectionexponent=30.0; 
  
  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: sphere: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      if (!strcmp(row,"center")) {
        sscanf(rest,"%lf , %lf , %lf",
            &SphereList[SphereCount].center[0],&SphereList[SphereCount].center[1],&SphereList[SphereCount].center[2]);
        SphereList[SphereCount].center[3]=1;
        dbgmessage("Load: sphere: center %.2f %.2f %.2f\n",
            SphereList[SphereCount].center[0],SphereList[SphereCount].center[1],SphereList[SphereCount].center[2]);
        break;
      }
      if (!strcmp(row,"color")) {
        sscanf(rest,"%lf , %lf , %lf",&SphereList[SphereCount].diffuseColor[0],
            &SphereList[SphereCount].diffuseColor[1],&SphereList[SphereCount].diffuseColor[2]);
        break;
      }
      if (!strcmp(row,"specular_color")) {
        sscanf(rest,"%lf , %lf , %lf",&SphereList[SphereCount].specularColor[0],
            &SphereList[SphereCount].specularColor[1],&SphereList[SphereCount].specularColor[2]);
        break;
      }
      if (!strcmp(row,"emissive_color")) {
        sscanf(rest,"%lf , %lf , %lf",&SphereList[SphereCount].emissiveColor[0],
            &SphereList[SphereCount].emissiveColor[1],&SphereList[SphereCount].emissiveColor[2]);
        break;
      }
      if (!strcmp(row,"reflection")) {
        sscanf(rest,"%lf , %lf , %lf",&SphereList[SphereCount].reflection[0],
            &SphereList[SphereCount].reflection[1],&SphereList[SphereCount].reflection[2]);
        dbgmessage("Load: sphere: reflection %.2f %.2f %.2f\n",SphereList[SphereCount].reflection[0],
            SphereList[SphereCount].reflection[1],SphereList[SphereCount].reflection[2]);
        break;
      }
      if (!strcmp(row,"radius")) {
        sscanf(rest,"%lf",&SphereList[SphereCount].radius);
        dbgmessage("Load: sphere: radius %.2f\n",SphereList[SphereCount].radius);
        break;
      }
      if (!strcmp(row,"refraction_index")) {
        sscanf(rest,"%lf",&SphereList[SphereCount].refractionindex);
        break;
      }
      if (!strcmp(row,"transparency")) {
        sscanf(rest,"%lf",&SphereList[SphereCount].transparency);
        dbgmessage("Load: sphere: transparency %.2f\n",SphereList[SphereCount].transparency);
        break;
      }
      if (!strcmp(row,"reflection_exponent")) {
        sscanf(rest,"%lf",&SphereList[SphereCount].reflectionexponent);
        dbgmessage("Load: sphere: reflection_exponent %.2f\n",SphereList[SphereCount].reflectionexponent);
        break;
      }
      dbgmessage("Load: sphere: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
  SphereCount++;
}

void LoadTriangle(void){
  char row[1000],*rest;
  int i,rl,v,ma_normaly=0;
  tVector v1,v2;
 
  // NASTAVIME DEFAULT HODNOTY FARBY A MATERIALU
  TriangleList[TriangleCount].ambientIntensity=0.15;
  TriangleList[TriangleCount].diffuseColor[0]=TriangleList[TriangleCount].diffuseColor[1]
    =TriangleList[TriangleCount].diffuseColor[2]=1.0;
  TriangleList[TriangleCount].specularColor[0]=TriangleList[TriangleCount].specularColor[1]
    =TriangleList[TriangleCount].specularColor[2]=0.0;
  TriangleList[TriangleCount].emissiveColor[0]=TriangleList[TriangleCount].emissiveColor[1]
    =TriangleList[TriangleCount].emissiveColor[2]=0.0;
  
  TriangleList[TriangleCount].reflection[AMBIENT]=0.15; 
  TriangleList[TriangleCount].reflection[DIFFUSE]=0.8; 
  TriangleList[TriangleCount].reflection[SPECULAR]=0.0; 
  
  TriangleList[TriangleCount].transparency=0.0; 
  TriangleList[TriangleCount].refractionindex=1.0; 
  TriangleList[TriangleCount].reflectionexponent=30.0; 
  
  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: triangle: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      if (sscanf(row,"vertex%d",&v)) {
        sscanf(rest,"%lf , %lf , %lf",&TriangleList[TriangleCount].vertex[v][0],
            &TriangleList[TriangleCount].vertex[v][1],&TriangleList[TriangleCount].vertex[v][2]);
        dbgmessage("Load: triangle: vertex %.2f %.2f %.2f\n",TriangleList[TriangleCount].vertex[v][0],
            TriangleList[TriangleCount].vertex[v][1],TriangleList[TriangleCount].vertex[v][2]);
        break;
      }
      if (sscanf(row,"normal%d",&v)) {
        sscanf(rest,"%lf , %lf , %lf",&TriangleList[TriangleCount].normal[v][0],
            &TriangleList[TriangleCount].normal[v][1],&TriangleList[TriangleCount].normal[v][2]);
        dbgmessage("Load: triangle: normal %.2f %.2f %.2f\n",TriangleList[TriangleCount].normal[v][0],
            TriangleList[TriangleCount].normal[v][1],TriangleList[TriangleCount].normal[v][2]);
        ma_normaly=1;
        break;
      }
      if (!strcmp(row,"color")) {
        sscanf(rest,"%lf , %lf , %lf",&TriangleList[TriangleCount].diffuseColor[0],
            &TriangleList[TriangleCount].diffuseColor[1],&TriangleList[TriangleCount].diffuseColor[2]);
        break;
      }
      if (!strcmp(row,"specular_color")) {
        sscanf(rest,"%lf , %lf , %lf",&TriangleList[TriangleCount].specularColor[0],
            &TriangleList[TriangleCount].specularColor[1],&TriangleList[TriangleCount].specularColor[2]);
        break;
      }
      if (!strcmp(row,"emissive_color")) {
        sscanf(rest,"%lf , %lf , %lf",&TriangleList[TriangleCount].emissiveColor[0],
            &TriangleList[TriangleCount].emissiveColor[1],&TriangleList[TriangleCount].emissiveColor[2]);
        break;
      }
      if (!strcmp(row,"reflection")) {
        sscanf(rest,"%lf , %lf , %lf",&TriangleList[TriangleCount].reflection[0],
            &TriangleList[TriangleCount].reflection[1],&TriangleList[TriangleCount].reflection[2]);
        dbgmessage("Load: triangle: reflection %.2f %.2f %.2f\n",TriangleList[TriangleCount].reflection[0],
            TriangleList[TriangleCount].reflection[1],TriangleList[TriangleCount].reflection[2]);
        break;
      }
      if (!strcmp(row,"refraction_index")) {
        sscanf(rest,"%lf",&SphereList[SphereCount].refractionindex);
        break;
      }
      if (!strcmp(row,"transparency")) {
        sscanf(rest,"%lf",&TriangleList[TriangleCount].transparency);
        dbgmessage("Load: triangle: transparency %.2f\n",TriangleList[TriangleCount].transparency);
        break;
      }
      if (!strcmp(row,"reflection_exponent")) {
        sscanf(rest,"%lf",&TriangleList[TriangleCount].reflectionexponent);
        dbgmessage("Load: triangle: reflection_exponent %.2f\n",TriangleList[TriangleCount].reflectionexponent);
        break;
      }
      dbgmessage("Load: triangle: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
  if (!ma_normaly){
    for (i=0;i<3;i++) { 
      v1[i]=TriangleList[TriangleCount].vertex[1][i]-TriangleList[TriangleCount].vertex[0][i];
      v2[i]=TriangleList[TriangleCount].vertex[2][i]-TriangleList[TriangleCount].vertex[0][i];
    }
    for (i=0;i<3;i++) {
      CrossProduct(v2,v1,&TriangleList[TriangleCount].normal[i]);
      Normalize(&TriangleList[TriangleCount].normal[i]);
      dbgmessage("New normal: %.3f %.3f %.3f\n",
          TriangleList[TriangleCount].normal[i][0],
          TriangleList[TriangleCount].normal[i][1],
          TriangleList[TriangleCount].normal[i][2]);
    }
  }
  TriangleCount++;
}

void LoadAmbientLight(void){
  char row[1000],*rest;
  int i,rl;
  
  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: sphere: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      
      if (!strcmp(row,"color")) {
        sscanf(rest,"%lf , %lf , %lf",&AmbientLight[0],&AmbientLight[1],&AmbientLight[2]);
        dbgmessage("Load: ambient_light_source: color %.2f %.2f %.2f\n",AmbientLight[0],AmbientLight[1],AmbientLight[2]);
        break;
      }
      
      dbgmessage("Load: ambient_light_source: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
}

void LoadPointLight(void){
  char row[1000],*rest;
  int i,rl;
  
  for (i=0;i<4;i++) PointLights[PointLightCount].direction[i]=0; 
  PointLights[PointLightCount].direction[2]=0;
  PointLights[PointLightCount].cosangle=-1.1;
  PointLights[PointLightCount].cosfalloff=-1.1;
  PointLights[PointLightCount].attenuation[0]=1.0;
  PointLights[PointLightCount].attenuation[1]=0.0;
  PointLights[PointLightCount].attenuation[2]=0.0;
  
  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: point_light_source: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      if (!strcmp(row,"location")) {
        sscanf(rest,"%lf , %lf , %lf",
            &PointLights[PointLightCount].location[0],
            &PointLights[PointLightCount].location[1],
            &PointLights[PointLightCount].location[2]);
        break;
      }
      if (!strcmp(row,"color")) {
        sscanf(rest,"%lf , %lf , %lf",
            &PointLights[PointLightCount].color[0],
            &PointLights[PointLightCount].color[1],
            &PointLights[PointLightCount].color[2]);
        break;
      }
      if (!strcmp(row,"attenuation")) {
        sscanf(rest,"%lf , %lf , %lf",
            &PointLights[PointLightCount].attenuation[0],
            &PointLights[PointLightCount].attenuation[1],
            &PointLights[PointLightCount].attenuation[2]);
        break;
      }
      dbgmessage("Load: point_light_source: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
  PointLightCount++;
}

void LoadSpotLight(void){
  char row[1000],*rest;
  int i,rl;
  double tmp;
  
  PointLights[PointLightCount].attenuation[0]=1.0;
  PointLights[PointLightCount].attenuation[1]=0.0;
  PointLights[PointLightCount].attenuation[2]=0.0;

  fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  while (strcmp(row,"end")) {
    while (1) { // trick, bude len raz, ale chceme vediet breakovat
      if ((row[0]=='/')&&(row[1]=='/')) break;
      rl=strlen(row); for (i=0;i<rl;i++) if (row[i]=='=') break;
      if (i==rl) { dbgmessage("Load: spot_light_source: I don't understand: '%s'\n",row); break; }
      rest=row+i+1; row[i]=0;
      if (!strcmp(row,"location")) {
        sscanf(rest,"%lf , %lf , %lf",
            &PointLights[PointLightCount].location[0],
            &PointLights[PointLightCount].location[1],
            &PointLights[PointLightCount].location[2]);
        break;
      }
      if (!strcmp(row,"color")) {
        sscanf(rest,"%lf , %lf , %lf",
            &PointLights[PointLightCount].color[0],
            &PointLights[PointLightCount].color[1],
            &PointLights[PointLightCount].color[2]);
        break;
      }
      if (!strcmp(row,"attenuation")) {
        sscanf(rest,"%lf , %lf , %lf",
            &PointLights[PointLightCount].attenuation[0],
            &PointLights[PointLightCount].attenuation[1],
            &PointLights[PointLightCount].attenuation[2]);
        break;
      }
      if (!strcmp(row,"look_at")) {
        sscanf(rest,"%lf , %lf , %lf",
            &PointLights[PointLightCount].direction[0],
            &PointLights[PointLightCount].direction[1],
            &PointLights[PointLightCount].direction[2]);
        for (i=0;i<3;i++) PointLights[PointLightCount].direction[i]-=PointLights[PointLightCount].location[i];
        Normalize(&PointLights[PointLightCount].direction);
        break;
      }
      if (!strcmp(row,"angle")) {
        sscanf(rest,"%lf",&tmp);
        PointLights[PointLightCount].cosangle=cos((tmp/180)*M_PI); 
        PointLights[PointLightCount].cosfalloff=PointLights[PointLightCount].cosangle; 
        dbgmessage("Load: spot_light_source: cos(angle) %.3f\n",PointLights[PointLightCount].cosangle);
        break;
      }
      if (!strcmp(row,"falloff")) {
        sscanf(rest,"%lf",&tmp);
        PointLights[PointLightCount].cosfalloff=cos((tmp/180)*M_PI); 
        dbgmessage("Load: spot_light_source: cos(falloff) %.3f\n",PointLights[PointLightCount].cosfalloff);
        break;
      }
      dbgmessage("Load: spot_light_source: Not implemented: '%s'\n",row);
      break; 
    }
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
  }
  PointLightCount++;
}

void LoadScene(void){
  int rl;
  char row[1000];
  
  if (!(f=fopen(infilename,"r"))) { message("Open failed.\n"); exit(1); }
  while (!feof(f)){
    fgets(row,1000,f); fscanf(f," "); rl=strlen(row); while (row[rl-1]<=32) row[--rl]=0;
    dbgmessage("Debug: loaded '%s'\n",row);
    if (strlen(row)==0) continue;
    if ((row[0]=='/')&&(row[1]=='/')) continue;
    if (!strcmp(row,"camera")) { LoadCamera(); continue; }
    if (!strcmp(row,"sphere")) { LoadSphere(); continue; }
    if (!strcmp(row,"triangle")) { LoadTriangle(); continue; }
    if (!strcmp(row,"ambient_light_source")) { LoadAmbientLight(); continue; }
    if (!strcmp(row,"point_light_source")) { LoadPointLight(); continue; }
    if (!strcmp(row,"spot_light_source")) { LoadSpotLight(); continue; }
    if (!strcmp(row,"eof")) break; 
    dbgmessage("Load: I don't understand: '%s'\n",row);
  }
  fclose(f);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   PRIESECNIKY                                             //
///////////////////////////////////////////////////////////////////////////////////////////////

void IntersectSphere(int gula, tPoint source, tVector smer, double *where){
  double A,B,C; // At^2 + Bt + C = 0
  double D,t1,t2;
  int k;

  A=0; for (k=0;k<3;k++) A+=smer[k]*smer[k];
  B=0; for (k=0;k<3;k++) B+=2*(smer[k]*(source[k]-SphereList[gula].center[k]));
  C=0; for (k=0;k<3;k++) C+=sqr(source[k]-SphereList[gula].center[k]); C-=sqr(SphereList[gula].radius);
  D=B*B - 4*A*C;
  if (D>=0) {
    D=sqrt(D);
    t1=(-B+D)/(2*A);
    t2=(-B-D)/(2*A);
    (*where)=INFTY;
    if (t1>EPS) (*where)=t1;
    if ((t2>EPS)&&(t2<(*where))) (*where)=t2;
  } else {
    (*where)=INFTY;
  }
}

void IntersectTriangle(int trojuh, tPoint source, tVector smer, double *where){
  tVector edge1,edge2,pvec,tvec,qvec,ssmer;
  int i;
  double det,inv,u,v,l;

  for (i=0;i<3;i++) { 
    edge1[i]=TriangleList[trojuh].vertex[1][i]-TriangleList[trojuh].vertex[0][i]; 
    edge2[i]=TriangleList[trojuh].vertex[2][i]-TriangleList[trojuh].vertex[0][i]; 
    ssmer[i]=smer[i];
  }
  l=Length(ssmer); Normalize(&ssmer);
  CrossProduct(ssmer,edge2,&pvec);
  det=DotProduct(edge1,pvec);
  if (fabs(det)<EPS) { (*where)=INFTY; return; } // sme v rovine trojuholnika
  inv=1/det;
  for (i=0;i<3;i++) tvec[i]=source[i]-TriangleList[trojuh].vertex[0][i]; 
  u=DotProduct(tvec,pvec)*inv;
  if ((u<EPS)||(u>1-EPS)) { (*where)=INFTY; return; } // sur. u je mimo
  CrossProduct(tvec,edge1,&qvec);
  v=DotProduct(ssmer,qvec)*inv;
  if ((v<EPS)||(u+v>1-EPS)) { (*where)=INFTY; return; } // sur. v je mimo
  (*where)=DotProduct(edge2,qvec)*inv/l;
}

void InterpolateTriangleNormal(int tr, tPoint bod, tVector *normal){
  double u,v;
  tVector smer1,smer2,vek;
  int i;
  for (i=0;i<3;i++) {
    vek[i]=bod[i]-TriangleList[tr].vertex[0][i];
    smer1[i]=TriangleList[tr].vertex[1][i]-TriangleList[tr].vertex[0][i];
    smer2[i]=TriangleList[tr].vertex[2][i]-TriangleList[tr].vertex[0][i];
  }
  if (fabs(smer1[0])+fabs(smer2[0]) < EPS) { smer1[0]=smer1[2]; smer2[0]=smer2[2]; }
  if (fabs(smer1[0]*smer2[1] - smer1[1]*smer2[0])<EPS) { smer1[1]=smer1[2]; smer2[1]=smer2[2]; }
  u=(vek[0]*smer2[1] - vek[1]*smer2[0]) / (smer1[0]*smer2[1] - smer1[1]*smer2[0]);
  v=(vek[0]*smer1[1] - vek[1]*smer1[0]) / (smer2[0]*smer1[1] - smer2[1]*smer1[0]);
//  dbgmessage("interpolation: u=%.3f v=%.3f\n",u,v);
  for (i=0;i<3;i++) 
    (*normal)[i]=u*TriangleList[tr].normal[1][i]+v*TriangleList[tr].normal[2][i]+(1-u-v)*TriangleList[tr].normal[0][i];
  
  Normalize(normal);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   TIENE a pod.                                            //
///////////////////////////////////////////////////////////////////////////////////////////////

void ShadowCoefficient(int svetlo, tPoint bod, tVector ksvetlu, tVector *coef){
  double where,uhol;
  int i;
  tVector pom1,pom2;

  // na zaciatku: svieti
  (*coef)[0]=(*coef)[1]=(*coef)[2]=1.0;
  
  // ak je to spotlight, pozrieme, ci naozaj
  for (i=0;i<3;i++) { pom1[i]=-PointLights[svetlo].direction[i]; pom2[i]=ksvetlu[i]; }
  Normalize(&pom1); Normalize(&pom2); uhol=DotProduct(pom1,pom2);
  if (uhol<PointLights[svetlo].cosfalloff) { (*coef)[0]=(*coef)[1]=(*coef)[2]=0.0; return; }
  if (uhol<PointLights[svetlo].cosangle) {
    (*coef)[0]*=(uhol-PointLights[svetlo].cosfalloff)/(PointLights[svetlo].cosangle-PointLights[svetlo].cosfalloff);
    (*coef)[1]*=(uhol-PointLights[svetlo].cosfalloff)/(PointLights[svetlo].cosangle-PointLights[svetlo].cosfalloff);
    (*coef)[2]*=(uhol-PointLights[svetlo].cosfalloff)/(PointLights[svetlo].cosangle-PointLights[svetlo].cosfalloff);
  }
  
  for (i=0;i<SphereCount;i++) { 
    IntersectSphere(i,bod,ksvetlu,&where); 
    if ((where>EPS) && (where<1-EPS)) {
      // luc ide na ceste od svetla tadeto
      (*coef)[0]*=SphereList[i].transparency*SphereList[i].diffuseColor[0];
      (*coef)[1]*=SphereList[i].transparency*SphereList[i].diffuseColor[1];
      (*coef)[2]*=SphereList[i].transparency*SphereList[i].diffuseColor[2];
//      if (((*coef)[0]+(*coef)[1]+(*coef)[2])<EPS) return;
    } 
  }
  for (i=0;i<TriangleCount;i++) { 
    IntersectTriangle(i,bod,ksvetlu,&where); 
    if ((where>EPS) && (where<1-EPS)) {
      // luc ide na ceste od svetla tadeto
      (*coef)[0]*=TriangleList[i].transparency*TriangleList[i].diffuseColor[0];
      (*coef)[1]*=TriangleList[i].transparency*TriangleList[i].diffuseColor[1];
      (*coef)[2]*=TriangleList[i].transparency*TriangleList[i].diffuseColor[2];
//      if (((*coef)[0]+(*coef)[1]+(*coef)[2])<EPS) return;
    } 
  }
}

double AttenuationFactor(int svetlo, tPoint bod){
  double dist,res;
  dist =sqr(bod[0]-PointLights[svetlo].location[0]);
  dist+=sqr(bod[1]-PointLights[svetlo].location[1]);
  dist+=sqr(bod[2]-PointLights[svetlo].location[2]);
  dist =sqrt(dist);
  res =PointLights[svetlo].attenuation[0];
  res+=PointLights[svetlo].attenuation[1]*dist;
  res+=PointLights[svetlo].attenuation[2]*sqr(dist);
  if (res<1) res=1;
  return (1/res);
}

int SnellRefraction(tVector normal,tVector *smer,double refractionindex, double cosangle) {
  int i;
  double haluz;
  if (cosangle>0) {
    for (i=0;i<3;i++) normal[i]=-normal[i];
  } else {
    cosangle=fabs(cosangle); refractionindex=1/refractionindex;
  }
  Normalize(smer);
  haluz=1-sqr(cosangle); haluz*=sqr(refractionindex); haluz=1-haluz;
  if (haluz<EPS) return 0;
  haluz=refractionindex*cosangle-sqrt(haluz);
  for (i=0;i<3;i++) (*smer)[i]=haluz*normal[i]+refractionindex*(*smer)[i];
  return 1;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   RAYTRACING                                              //
///////////////////////////////////////////////////////////////////////////////////////////////

void CastRay(tPoint source, tVector smer, tVector* farba, int vnorenie){
  double nearest,where,angle,tmp,att;
  int i,co,typ,sv;
  tVector normal,light,tolight,coef,reflect,nsmer,rfarba;
  tPoint intersection;
  
  nearest=INFTY;
  for (i=0;i<SphereCount;i++) 
    { IntersectSphere(i,source,smer,&where);  if ((where>EPS) && (where<nearest)) { nearest=where; co=i; typ=1; } }
  for (i=0;i<TriangleCount;i++) { 
    IntersectTriangle(i,source,smer,&where);  
    if ((where>EPS) && (where<nearest)) { nearest=where; co=i; typ=2; } 
  }

  
  if (INFTY-nearest>EPS) {
    // nasli sme priesecnik
    
    if (typ==1) { // gula
      for (i=0;i<3;i++) (*farba)[i]=0;
      // ambient zlozka svetla
      for (i=0;i<3;i++) 
        (*farba)[i]+=AmbientLight[i]*SphereList[co].reflection[AMBIENT]*SphereList[co].diffuseColor[i]; 
      // diffuse a specular zlozka svetla
      for (i=0;i<3;i++) normal[i]=source[i]+smer[i]*nearest-SphereList[co].center[i]; Normalize(&normal);
      for (sv=0;sv<PointLightCount;sv++){
        for (i=0;i<3;i++) light[i]=PointLights[sv].location[i]-(source[i]+smer[i]*nearest); Normalize(&light);
        angle=DotProduct(normal,light);
        for (i=0;i<3;i++) { reflect[i]=2*normal[i]*angle-light[i]; nsmer[i]=-smer[i]; } 
        Normalize(&reflect); Normalize(&nsmer);
        tmp=pow(DotProduct(nsmer,reflect),SphereList[co].reflectionexponent);
        if (TWOSIDED_LIGHTING) angle=fabs(angle);
        if (angle>0) {
          // sme na osvetlenej casti, ak nie sme v tieni
          for (i=0;i<4;i++) intersection[i]=source[i]+nearest*smer[i];
          for (i=0;i<4;i++) tolight[i]=PointLights[sv].location[i]-intersection[i];
          for (i=0;i<4;i++) intersection[i]+=2*EPS*tolight[i];
          ShadowCoefficient(sv,intersection,tolight,&coef);
          att=AttenuationFactor(sv,intersection);
          att*=(1-SphereList[co].transparency);
          for (i=0;i<3;i++) (*farba)[i]+=att*coef[i]*PointLights[sv].color[i]*SphereList[co].diffuseColor[i]*angle; 
          // for (i=0;i<3;i++) (*farba)[i]+=att*coef[i]*PointLights[sv].color[i]*SphereList[co].specularColor[i]*tmp; 
          for (i=0;i<3;i++) (*farba)[i]+=att*coef[i]*PointLights[sv].color[i]*SphereList[co].reflection[SPECULAR]*tmp; 
        }
      }
      // emissive zlozka (tfuj)
      for (i=0;i<3;i++) (*farba)[i]+=SphereList[co].emissiveColor[i];
      // priratame odrazeny luc
//      if ((SphereList[co].reflection[SPECULAR]>EPS) && (vnorenie<HLBKA_VNORENIA)) {
      if ((SphereList[co].specularColor[0]+SphereList[co].specularColor[1]+SphereList[co].specularColor[2])>EPS)
      if (vnorenie<HLBKA_VNORENIA) {
        for (i=0;i<3;i++) nsmer[i]=-smer[i]; Normalize(&nsmer);     angle=DotProduct(normal,nsmer);
        for (i=0;i<3;i++) reflect[i]=(2*normal[i]*angle-nsmer[i]);  Normalize(&reflect);
        for (i=0;i<4;i++) intersection[i]=source[i]+nearest*smer[i];
        CastRay(intersection,reflect,&rfarba,vnorenie+1);
        for (i=0;i<3;i++) (*farba)[i]+=rfarba[i]*SphereList[co].specularColor[i]; 
      }
      // priratame zlomeny luc
      if ((SphereList[co].transparency>EPS) && (vnorenie<HLBKA_VNORENIA)) {
        for (i=0;i<4;i++) intersection[i]=source[i]+nearest*smer[i];
        for (i=0;i<4;i++) nsmer[i]=smer[i]; Normalize(&nsmer);
        for (i=0;i<3;i++) normal[i]=source[i]+smer[i]*nearest-SphereList[co].center[i]; Normalize(&normal);
        angle=DotProduct(normal,nsmer);
        if (SnellRefraction(normal,&nsmer,SphereList[co].refractionindex,angle)) {
          // ak vobec nastava lom
          CastRay(intersection,nsmer,&rfarba,vnorenie+1);
          for (i=0;i<3;i++) (*farba)[i]+=rfarba[i]*SphereList[co].transparency*SphereList[co].diffuseColor[i]; 
        }
      }
    } // typ=gula
    
    if (typ==2) { // trojuholnik
      for (i=0;i<3;i++) (*farba)[i]=0;
      // ambient zlozka svetla
      for (i=0;i<3;i++) 
        (*farba)[i]+=AmbientLight[i]*TriangleList[co].reflection[AMBIENT]*TriangleList[co].diffuseColor[i]; 
      // diffuse a specular zlozka svetla
      for (i=0;i<4;i++) intersection[i]=source[i]+nearest*smer[i];
      InterpolateTriangleNormal(co,intersection,&normal);
//      dbgmessage("Interpoovana normala: %.3f %.3f %.3f\n",normal[0],normal[1],normal[2]);
      for (sv=0;sv<PointLightCount;sv++){
        for (i=0;i<3;i++) light[i]=PointLights[sv].location[i]-(source[i]+smer[i]*nearest); Normalize(&light);
        angle=DotProduct(normal,light);
        for (i=0;i<3;i++) { reflect[i]=2*normal[i]*angle-light[i]; nsmer[i]=-smer[i]; } 
        Normalize(&reflect); Normalize(&nsmer);
        tmp=pow(DotProduct(nsmer,reflect),TriangleList[co].reflectionexponent);
        if (TWOSIDED_LIGHTING) angle=fabs(angle);
        if (angle>0) {
          // sme na osvetlenej casti, ak nie sme v tieni
          for (i=0;i<4;i++) intersection[i]=source[i]+nearest*smer[i];
          for (i=0;i<4;i++) tolight[i]=PointLights[sv].location[i]-intersection[i];
          for (i=0;i<4;i++) intersection[i]+=2*EPS*tolight[i];
          ShadowCoefficient(sv,intersection,tolight,&coef);
          att=AttenuationFactor(sv,intersection);
          for (i=0;i<3;i++) (*farba)[i]+=att*coef[i]*PointLights[sv].color[i]*TriangleList[co].diffuseColor[i]*angle; 
          for (i=0;i<3;i++)(*farba)[i]+=att*coef[i]*PointLights[sv].color[i]*TriangleList[co].reflection[SPECULAR]*tmp; 
        }
      }
      // emissive zlozka (tfuj)
      for (i=0;i<3;i++) (*farba)[i]+=TriangleList[co].emissiveColor[i];
      // priratame odrazeny luc
//      if ((TriangleList[co].reflection[SPECULAR]>EPS) && (vnorenie<HLBKA_VNORENIA)) {
      if ((TriangleList[co].specularColor[0]+TriangleList[co].specularColor[1]+TriangleList[co].specularColor[2])>EPS)
      if (vnorenie<HLBKA_VNORENIA) {
        for (i=0;i<3;i++) nsmer[i]=-smer[i]; Normalize(&nsmer);     angle=DotProduct(normal,nsmer);
        for (i=0;i<3;i++) reflect[i]=(2*normal[i]*angle-nsmer[i]);  Normalize(&reflect);
        for (i=0;i<4;i++) intersection[i]=source[i]+nearest*smer[i];
        CastRay(intersection,reflect,&rfarba,vnorenie+1);
        for (i=0;i<3;i++) (*farba)[i]+=rfarba[i]*TriangleList[co].specularColor[i]; 
      }
      // priratame zlomeny luc
      if ((TriangleList[co].transparency>EPS) && (vnorenie<HLBKA_VNORENIA)) {
        for (i=0;i<4;i++) intersection[i]=source[i]+nearest*smer[i];
        for (i=0;i<4;i++) nsmer[i]=smer[i]; Normalize(&nsmer);
        InterpolateTriangleNormal(co,intersection,&normal); 
        angle=DotProduct(normal,nsmer);
        if (SnellRefraction(normal,&nsmer,TriangleList[co].refractionindex,angle)) {
          // ak vobec nastava lom
          CastRay(intersection,smer,&rfarba,vnorenie+1);
          for (i=0;i<3;i++) (*farba)[i]+=rfarba[i]*TriangleList[co].transparency*TriangleList[co].diffuseColor[i]; 
        }
      }
    } // typ=trojuholnik
    
    // upravime
    for (i=0;i<3;i++) if ((*farba)[i]<0.0) (*farba)[i]=0.0; 
    for (i=0;i<3;i++) if ((*farba)[i]>1.0) (*farba)[i]=1.0; 
  } else {
    (*farba)[0]=(*farba)[1]=(*farba)[2]=0.0;
  }
}

void RayTrace(void){
  FILE *fout;
  tVector farba,smer;
  int i,j,k,l;
  double x,y;

  fout=fopen("output.bmp","w");
  fprintf(fout,"%c%c%c%c",0x42,0x4D,0x36,0x10);
  fprintf(fout,"%c%c%c%c",0x0E,0x00,0x00,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x36,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x28,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,XSIZE%256,XSIZE/256);
  
  fprintf(fout,"%c%c%c%c",0x00,0x00,YSIZE%256,YSIZE/256);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x01,0x00);
  fprintf(fout,"%c%c%c%c",0x18,0x00,0x00,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x00,0x10);
  fprintf(fout,"%c%c%c%c",0x0E,0x00,0x6D,0x0B);
  
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x6D,0x0B);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x00,0x00);
  fprintf(fout,"%c%c%c%c",0x00,0x00,0x00,0x00);
  fprintf(fout,"%c%c",0x00,0x00);

  for (j=0;j<YSIZE;j++) { 
    for (i=0;i<XSIZE;i++) {
      // spocitame si bod sveta, cez ktory ideme 
      x=2*i-XSIZE; y=2*j-YSIZE; x/=XSIZE; y/=YSIZE;
      for (k=0;k<3;k++) smer[k]=Kamera.direction[k]+x*Kamera.right[k]+y*Kamera.up[k];
      //message("smer: %.2f %.2f %.2f\n",smer[0],smer[1],smer[2]);
      CastRay(Kamera.location,smer,&farba,1);
      for (l=2;l>=0;l--) { k=(int)(farba[l]*255); putc((char)k,fout); }
    }
    message("Rows done: %d\r",j+1); fflush(stdout);
  }
  fclose(fout);
}

///////////////////////////////////////////////////////////////////////////////////////////////
//                                   MAIN                                                    //
///////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv){
  int i,pom;

  infilename[0]=0;
  for (i=1;i<argc;i++){
    if (!strcmp(argv[i],"-v")) { VERBOSE=1; continue; }
    if (!strcmp(argv[i],"-d")) { DEBUG=1; continue; }
    if (sscanf(argv[i],"-x%d",&pom)) { XSIZE=pom; continue; }
    if (sscanf(argv[i],"-y%d",&pom)) { YSIZE=pom; continue; }
    if (sscanf(argv[i],"-depth%d",&pom)) { HLBKA_VNORENIA=pom; continue; }
    if (!strcmp(argv[i],"-badnormals")) { TWOSIDED_LIGHTING=1; continue; }
    if (!strlen(infilename)) strcpy(infilename,argv[i]);
  }

  message("Loading scene from file '%s'.\n",infilename);
  LoadScene();
  RayTrace();

  return 0;
}
