#ifndef CONTROLS_H
#define CONTROLS_H

#define AMBIENT 0
#define POINT 1
#define SPOT 2
#define DIFFUSE 1
#define SPECULAR 2
#define VELA 10000

#include <qcolordialog.h>
#include <qfiledialog.h>
#include <qinputdialog.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qlineedit.h>
#include <qlistbox.h>
#include <qmessagebox.h>
#include <qpushbutton.h>
#include <qwidget.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

// veci z raytracera

typedef double tMatrix[4][4];
typedef double tPoint[4];
typedef double tVector[4];

typedef struct {
  tPoint vertex[3];
  tVector normal[3];
//  QColor color;
  double ambientIntensity;
  QColor diffuseColor,specularColor,emissiveColor;
  tVector reflection;
  double reflectionexponent,refractionindex,transparency;
  int deleted;
} tTriangle;

typedef struct {
  tPoint center;
  double radius;
//  QColor color;
  double ambientIntensity;
  QColor diffuseColor,specularColor,emissiveColor;
  tVector reflection;
  double reflectionexponent,refractionindex,transparency;
  int deleted;
} tSphere;

typedef struct {
  int type;
  tPoint location;
  tVector direction;
  QColor color;
  double angle, falloff;
  tVector attenuation;
  int deleted;
} tLight;

typedef struct {
  tPoint location;
  tVector direction,right,up;
  double angle;
} tCamera;

// veci z qt

class PleaseWait : public QWidget
{
    Q_OBJECT

public:
    PleaseWait( QWidget* parent, const char* name);
    ~PleaseWait();
signals:
public slots:
private: 
};

class Controls : public QWidget
{
    Q_OBJECT

public:

    Controls( QWidget* parent, const char* name);
    ~Controls();

    QListBox *lightList, *triangleList, *sphereList;
    QLabel *lightPokec, *trianglePokec, *spherePokec, *cameraPokec;
    int triangleCount, lightCount, sphereCount;
    
    tLight lights[VELA];
    tTriangle triangles[VELA];
    tSphere spheres[VELA];
    tCamera Kamera;

signals:
  
    void dieSIGNAL();
    void repaintSIGNAL();

public slots:
 
    void loadVRMLSLOT();
    void loadSceneSLOT();
    void saveSceneSLOT();

    void changeCameraLocationSLOT();
    void changeCameraLookAtSLOT();
    void changeCameraUpSLOT();
    
    void raytraceSLOT();
    void quitButtonPushedSLOT();
    
    void changeLightColorSLOT();
    void changeLightLocationSLOT();
    void changeLightLookAtSLOT();
    void changeLightAngleSLOT();
  
    void deleteLightSLOT();
    void deleteSphereSLOT();
    void deleteTriangleSLOT();
    
    void addAmbientLightSLOT();
    void addPointLightSLOT();
    void addSpotLightSLOT();

    void describeLightSLOT();
    void describeCameraSLOT();
    void describeTriangleSLOT();
    void describeSphereSLOT();

private: 

    FILE *f;
    QString currentFileName;

    int LoadScene(QString infilename);
    void LoadSpotLight();
    void LoadPointLight();
    void LoadAmbientLight();
    void LoadTriangle();
    void LoadSphere();
    void LoadCamera();
    int SaveScene(QString infilename);

    void addLight(int);
    int getIntVector(QString,QString,int*,int*,int*);
    int getDouble(QString,QString,double*);
    int getDoubleVector(QString,QString,double*,double*,double*);
    
};

#endif // CONTROLS_H
