#include <qgl.h>
#include "glcanvas.h"
#include "controls.h"
#include <stdio.h>

GLCanvas::GLCanvas( QWidget* parent, const char* name, Controls *myControls) : QGLWidget( parent, name ) 
{ 
  theControls=myControls; 
}

GLCanvas::~GLCanvas()
{
}


void CreateSphere(tPoint c,double r,int n)
{
   int i,j;
   double theta1,theta2,theta3;
   double TWOPI=2*3.14159265358979323846,PID2=3.14159265358979323846/2;
   tPoint e,p;
   if (r < 0) r = -r; if (n < 0) n = -n;
   if (n < 4 || r <= 0) { glBegin(GL_POINTS); glVertex3f(c[0],c[1],c[2]); glEnd(); return; }
   for (j=0;j<n/2;j++) {
      theta1 = j * TWOPI / n - PID2;
      theta2 = (j + 1) * TWOPI / n - PID2;

      glBegin(GL_QUAD_STRIP);
      for (i=0;i<=n;i++) {
         theta3 = i * TWOPI / n;

         e[0] = cos(theta2) * cos(theta3);  e[1] = sin(theta2);      e[2] = cos(theta2) * sin(theta3);
         p[0] = c[0] + r * e[0];            p[1] = c[1] + r * e[1];  p[2] = c[2] + r * e[2];
         glNormal3d(e[0],e[1],e[2]);
         glTexCoord2f(i/(double)n,2*(j+1)/(double)n);
         glVertex3d(p[0],p[1],p[2]);

         e[0] = cos(theta1) * cos(theta3);  e[1] = sin(theta1);      e[2] = cos(theta1) * sin(theta3);
         p[0] = c[0] + r * e[0];            p[1] = c[1] + r * e[1];  p[2] = c[2] + r * e[2];
         glNormal3d(e[0],e[1],e[2]);
         glTexCoord2f(i/(double)n,2*j/(double)n);
         glVertex3d(p[0],p[1],p[2]);
      }
      glEnd();
   }
}


double aaaaaDotProduct(tVector v1, tVector v2) { return (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]); }
double aaaaaLength(tVector v1) { return sqrt(aaaaaDotProduct(v1,v1)); }

void GLCanvas::paintGL()
{

   double alpha = atan2(aaaaaLength(theControls->Kamera.up),aaaaaLength(theControls->Kamera.direction));
   alpha *= (180/M_PI);
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity ();
   gluPerspective(2*alpha,4/3.0,1.5,40.0);
   glMatrixMode (GL_MODELVIEW);

   
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glColor3d (1.0, 1.0, 1.0);
  glLoadIdentity ();
  glScalef(-1.0, 1.0, 1.0);

//  glScalef(0.3,0.3,0.3);
//  gluLookAt(-10.0, 5.0, 7.0, -9.0, 4.9, 7.4, 0.0, 1.0, 0.0);
  gluLookAt(
    theControls->Kamera.location[0],  
    theControls->Kamera.location[1],  
    theControls->Kamera.location[2],  
    theControls->Kamera.location[0]+theControls->Kamera.direction[0],  
    theControls->Kamera.location[1]+theControls->Kamera.direction[1],  
    theControls->Kamera.location[2]+theControls->Kamera.direction[2],  
    theControls->Kamera.up[0],  
    theControls->Kamera.up[1],  
    theControls->Kamera.up[2]);
//  glScalef (1.0, 2.0, 1.0);      /* modeling transformation */
//  glRotatef (alpha, 1, 1, 1);

  int R,G,B,uz=0;

  static float lmodel_ambient[] = {0.0, 0.0, 0.0, 0.0};
  static float lmodel_twoside[] = {GL_FALSE};
  static float lmodel_local[] = {GL_FALSE};
  glLightModelfv(GL_LIGHT_MODEL_LOCAL_VIEWER, lmodel_local);
  glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
  glEnable(GL_LIGHTING);

  static float light_color[] = {0.1, 0.1, 0.1, 1.0};
  static float light_position[] = {0.8660254, 0.5, 1, 0};
  for (int i=0;i<theControls->lightCount;i++) {
    if (theControls->lights[i].deleted) { /* glDisable(GL_LIGHT0 +uz); */ continue; }
    switch (theControls->lights[i].type) {
      case AMBIENT:
        theControls->lights[i].color.rgb(&R,&G,&B);
        light_color[0]=R/255.0; light_color[1]=G/255.0; light_color[2]=B/255.0;
        glLightfv(GL_LIGHT0 +uz, GL_AMBIENT, light_color);
        break;
    }
    glEnable(GL_LIGHT0 +uz);
    uz++;
  }
  for (int i=0;i<theControls->lightCount;i++) {
    if (theControls->lights[i].deleted) { /* glDisable(GL_LIGHT0 +uz); */ continue; }
    switch (theControls->lights[i].type) {
      case POINT:
        theControls->lights[i].color.rgb(&R,&G,&B);
        light_color[0]=R/255.0; light_color[1]=G/255.0; light_color[2]=B/255.0;
        glLightfv(GL_LIGHT0 +uz, GL_DIFFUSE, light_color);
        for (int hh=0;hh<3;hh++) light_position[hh]=(float) theControls->lights[i].location[hh];
        light_position[3]=1.0;
        glLightfv(GL_LIGHT0 +uz, GL_POSITION, light_position);
        break;
      case SPOT:
        theControls->lights[i].color.rgb(&R,&G,&B);
        light_color[0]=R/255.0; light_color[1]=G/255.0; light_color[2]=B/255.0;
        glLightfv(GL_LIGHT0 +uz, GL_DIFFUSE, light_color);
        for (int hh=0;hh<3;hh++) light_position[hh]=(float) theControls->lights[i].location[hh];
        light_position[3]=1.0;
        glLightfv(GL_LIGHT0 +uz, GL_POSITION, light_position);
        glLightf(GL_LIGHT0 +uz,GL_SPOT_CUTOFF, (float)theControls->lights[i].angle);
        for (int hh=0;hh<3;hh++) light_position[hh]=(float) theControls->lights[i].direction[hh];
        light_position[3]=0.0;
        glLightfv(GL_LIGHT0 +uz,GL_SPOT_DIRECTION, light_position);
        glLightf(GL_LIGHT0 +uz,GL_SPOT_EXPONENT, 0.0);
        break;
    }
    glEnable(GL_LIGHT0 +uz);
    uz++;
  }
  for (;uz<=20;uz++) glDisable(GL_LIGHT0 +uz);

  glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  glBegin(GL_TRIANGLES);
    for (int i=0;i<theControls->triangleCount;i++) if (!(theControls->triangles[i].deleted)) {
      
      theControls->triangles[i].diffuseColor.rgb(&R,&G,&B);
      glColor3d(R/255.0,G/255.0,B/255.0);
     
      float pom[4]; pom[3]=1.0;
      
      pom[0]=pom[1]=pom[2]=theControls->triangles[i].reflection[AMBIENT];
      pom[0]*=R/255.0; pom[1]*=G/255.0; pom[2]*=B/255.0;
      glMaterialfv(GL_FRONT, GL_AMBIENT, pom);
      
      pom[0]=pom[1]=pom[2]=theControls->triangles[i].reflection[DIFFUSE];
      pom[0]*=R/255.0; pom[1]*=G/255.0; pom[2]*=B/255.0;
      glMaterialfv(GL_FRONT, GL_DIFFUSE, pom);
      
      glMaterialf(GL_FRONT, GL_SHININESS, theControls->triangles[i].reflectionexponent);
      pom[0]=pom[1]=pom[2]=theControls->triangles[i].reflection[SPECULAR];
      pom[0]*=R/255.0; pom[1]*=G/255.0; pom[2]*=B/255.0;
      glMaterialfv(GL_FRONT, GL_SPECULAR, pom);
    
      glNormal3dv(theControls->triangles[i].normal[0]);
      glVertex3dv(theControls->triangles[i].vertex[0]);
      glNormal3dv(theControls->triangles[i].normal[1]);
      glVertex3dv(theControls->triangles[i].vertex[1]);
      glNormal3dv(theControls->triangles[i].normal[2]);
      glVertex3dv(theControls->triangles[i].vertex[2]);
    }
  glEnd();

  for (int i=0;i<theControls->sphereCount;i++)  if (!(theControls->spheres[i].deleted)) {
    theControls->spheres[i].diffuseColor.rgb(&R,&G,&B);
    glColor3d(R/255.0,G/255.0,B/255.0);
    CreateSphere(theControls->spheres[i].center,theControls->spheres[i].radius,20);
  }

  glFlush();
}

void GLCanvas::initializeGL()
{
   // Init the GL Window Here.

   glClearColor(0, 0, 0, 0);
   glDepthFunc(GL_LESS); glEnable(GL_DEPTH_TEST); glClearDepth(1.0);

   glShadeModel(GL_SMOOTH);
}

void GLCanvas::resizeGL( int w, int h )
{
   // GL Resize Code
   double alpha = atan2(aaaaaLength(theControls->Kamera.up),aaaaaLength(theControls->Kamera.direction));
   alpha *= (180/M_PI);
//   printf("%.3f\n",alpha);

   glViewport (0, 0, (GLsizei) w, (GLsizei) h);
   glMatrixMode (GL_PROJECTION);
   glLoadIdentity ();
//   glFrustum (-1.0, 1.0, -1.0, 1.0, 1.5, 20.0);
//   glFrustum (-1.0, 1.0, -1.0, 1.0, 1.5, 30.0);
   gluPerspective(2*alpha,4/3.0,1.5,40000.0);
   glMatrixMode (GL_MODELVIEW);

}

void GLCanvas::repaintSLOT()
{
//   alpha = (alpha + 30) % 360;
   updateGL();
}

