//
// OpenVRML-Triangulate
//
// Copyright (C) 2003  Vlado Koutny
// Based on OpenVRML-GL
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 

/* usage:

    tri_VRML vrml;
    tri_Scene scene;
    
    vrml=tri_Open("ruza.wrl");
    tri_SetDivisions(vrml,15,10,12);
    scene=tri_Convert(vrml);
    tri_Close(vrml);
    
    
    compile with: gcc -lopenvrml -lopenvrml-tri your_app.c -o your_app
*/

#ifndef __TRIVRML_H
#define __TRIVRML_H

// #include <GL/glut.h>

typedef struct {
    double x,y,z;		// coordinates
    double nx,ny,nz;		// normal vector
} tri_Point;
  
typedef tri_Point tri_3Point[3];

typedef struct {
    double ambientIntensity;	// ambientna farba (this*diffuseColor*LightAmbientIntensity)
    double diffuseColor[3];	// difuzna farba
    double emissiveColor[3];	// svietivost objektu sameho
    double shininess;		// specular exponent (phong_n = 128*this)
    double specularColor[3];	// specular color (filtruje farbu odlesku svetla)
    double transparency;	// 1=transparent, 0=opaque
    int attenuated;		// ci sa ma pouzivat attenuation svetiel na tento material (vhodne pre background)
    
    double indexLomu;		// !!! vzdy 1 - VRML nema takuto polozku (ale mozete si ju nastavit vo vasej app)
} tri_Material;

typedef struct {
    double Pos[3];		// position
    double Dir[3];		// direction
    double Up[3];		// up-vector
    double Left[3];		// left-vector
    double Angle;		// viewing angle
    double zNear, zFar;		// near/far clipping plane (for gluPerspective)
} tri_Camera;

typedef struct {
    int type;			// 0=dir, 1=point, 2=spot
    double ambient;		// all
    double intensity;		// all
    double pos[3];		// point,spot
    double dir[3];		// dir,spot
    double color[3];		// all
    double attn[3];		// point,spot
    double radius;		// point,spot
    double beamWidth;		// spot
    double cutOffAngle;		// spot
} tri_Light;

typedef struct {
    tri_Material Material;	// object material
    int nTriangles;		// number of triangles
    tri_3Point *Triangle;	// points of all triangles
} tri_Object;

typedef struct {
    tri_Camera Camera;		// camera (position, look-at, ...)
    
    int nLights;		// number of lights in the scene
    tri_Light *Light;		// lights (position, color, ...)
    
    int nObjects;		// number of objects
    tri_Object *Object;		// object description (material, list of triangles, ...)
} tri_Scene;


typedef struct {		// VRML identificator
    void * vrmlscene;		// loaded vrml object
    void * vrmlobject;		// viewer object
    tri_Scene scene;
} tri_VRML;


// Load VRML file, do not convert yet
// returns pointer to vrml object created internally (for further referencing)
tri_VRML tri_Open(const char *name);

// Set various parameteres: Sphere, Cylinder, Cone divisions
void tri_SetDivisions(tri_VRML id, int sphere, int cylinder, int cone);

// Convert scene into triangles
tri_Scene tri_Convert(tri_VRML id);

// free allocated memory, ...
void tri_Close(tri_VRML id);

#endif /* __TRIVRML_H */

